# -*- coding: utf-8 -*-
'''
$RCSfile: $
$Revision: 1301 $
$Author: j.holetschek $
$Date: 2013-07-04 15:56:34 +0200 (Do, 04. Jul 2013) $
'''

import sys, os, zipfile, xml.dom.minidom, shutil, subprocess, time, codecs  #traceback
from os.path import join, exists, getsize, splitext
from datetime import datetime
import xml.etree.ElementTree as ET

# ***** include the biocase.lib directory in the python sys path for importing *****
sys.path.insert(0, join(os.path.dirname(__file__), os.pardir, os.pardir))

import simplejson as json
import biocase.configuration, logging.handlers
from biocase.datasources import Datasource
from biocase.wrapper.protocol.BioCASe.biocase_request import RequestClass
from biocase.wrapper.protocol.inventory.inventory_protocol import ProtocolClass as InventoryProtocolClass
from biocase.wrapper.operations import getOperationsObjectForDBMod
from biocase.archive.general import drop_archives, DWCA_SCHEMA_NAME
from biocase.tools.xmlutils import cleanupNamespaces, encode_for_xml  #prettyprint
from biocase.wrapper.sql.base import sqlBase
from general import NS_ABCD206, NS_HISPID5, list_archives_for_schema, create_xmlarchive_filename, inventory_add, fmt


BIOCASE_SEARCH_REQ = """<?xml version='1.0' encoding='UTF-8'?>
    <request xmlns='http://www.biocase.org/schemas/protocol/1.3'>
        <header><type>search</type></header>
        <search>
            <requestFormat>%s</requestFormat>
            <responseFormat start='0' limit='%i'>%s</responseFormat>
            <count>false</count>
            <filter>%s</filter>
        </search>
    </request>"""

DATASET_FILTER_TEMPLATE = "<equals path='/DataSets/DataSet/Metadata/Description/Representation/Title'>%s</equals>"


# Get arguments
xmla = sys.argv[1]
dwca = sys.argv[2]
dsa = sys.argv[3]
schema = sys.argv[4]
transformer = sys.argv[5]
datasets = json.loads(sys.argv[6])
filenames = json.loads(sys.argv[7])

# Load BioCASe cfg
cfg = biocase.configuration.Cfg()
dsaObj = Datasource(dsa)
archivePath = join(cfg.archiveWorkLocator, dsa)
downloadPath = join(cfg.archiveDownloadLocator, dsa)
logFile = join(archivePath, schema + ".log")
procFilename = join(archivePath, schema + ".proc")
configFile = join(archivePath, schema + ".config.xml")
proc_dwca = join(archivePath, DWCA_SCHEMA_NAME + '.proc')
tmpPath = join (archivePath, 'tmp')
dupFile = join (tmpPath, 'duplicate.txt')

# Create logger for the libs, with a handler that does nothing (so not logging)
logging.getLogger().addHandler(logging.handlers.MemoryHandler(0))


############################################################################################################
### Helper functions
############################################################################################################

def log(message, tag='INFO'):
    """
    Writes the message to the log file, with a leading header
       message: string to be written (can be multi-line string). None will print a line feed.
       tag: type of log entry. default is INFO
    """
    l = codecs.open(logFile, 'a', encoding='utf-8')
    if message is None:
        l.write('\n')
    else:
        ts = datetime.now()
        for line in message.splitlines():
            l.write('%-5s %02i-%02i %02i:%02i:%02i,%03i - %s\n' % (tag, ts.day, ts.month, ts.hour, ts.minute, ts. second, ts.microsecond / 1000, line))
    l.close()

def do_archive(opObj, limit, schemans, dataset=None, use_filter=False):
    """
    Create an XML archive for the specified schema.
       limit: paging size
       schemans: namespace of schema to be archived
       dataset: if set, archive only this ABCD dataset. default is None (for dataset-unaware archiving)
       use_filter: if set, use filter to allow dataset-aware filtering. this allows to skip the filter for all ABCD datasources
            that publish only one dataset (performance!)
    """

    # If dataset parameter is None, archiving will be dataset-unaware
    log(None)
    if dataset is None:
        archiveFilename = schema + '.zip'
        archiveFilenameFull = join(archivePath, archiveFilename)
        reqFilter = ''
    else:
        log('Archiving <em>%s</em>' % (dataset))
        archiveFilename = create_xmlarchive_filename(dsa, archivePath, downloadPath, dataset, schema, schemans)
        #archiveFilename = create_xmlarchive_filename(archivePath, downloadPath, dataset, schema)
        archiveFilenameFull = join(archivePath, archiveFilename)
        if use_filter:
            reqFilter = DATASET_FILTER_TEMPLATE % (encode_for_xml(dataset))
            #log(reqFilter, "DEBUG")
        else:
            reqFilter = ''

    # Init objects
    requestObj = RequestClass(dsaObj)
    reqString = BIOCASE_SEARCH_REQ % (schemans, limit, schemans, reqFilter)
    requestObj.parseRequest({'request_string': reqString})
    #log(str(requestObj))
    # Additonal check: If request could not be parsed, request type will be set to capabilities
    if requestObj.type == 'capabilities':
        log('Error parsing the request string for dataset <em>%s</em>. Skipping dataset.' % (dataset), "ERROR")
        return None, 0, 0, 0
    zipFile = zipfile.ZipFile(archiveFilenameFull, "w", zipfile.ZIP_DEFLATED, True)

    # Page through dataset/datasource
    offset, recordsArchived, dropped, page = 0, 0, 0, 1
    finished, cancelled = False, False
    while not (finished or cancelled):
        # Do the request
        log("Retrieving records %s - %s" % fmt(offset, offset + limit))
        (resultET, status) = opObj.search(requestObj)
        # we need to manually correct the total count if OFFSET clause is used (add the OFFSET to totalSearchHits)
        # usually this is done by biocase_response.postprocess, but we skip that
        if sqlBase.templates['select-limit-offset']:
            status.total += status.start
        log("   %s records archived (%s dropped, total hits: %s)" % fmt(status.count, status.drop, status.total))
        recordsArchived += status.count
        dropped += status.drop

        # Store the result (no prettyprinting)
        if resultET is not None:
            cleanupNamespaces(resultET)
            # prettyprint(resultET, stream=l, encoding='UTF-8')
            zipFile.writestr('response.%05i.xml' % (page), ET.tostring(resultET, 'UTF-8'))

        # Do we need to adjust paging size?
        if status.count + status.drop < limit:
            if offset + status.count + status.drop < status.total:
                log("Adjusting paging size to " + str(status.count + status.drop))
                limit = status.count + status.drop

        # Did we get all records?
        if offset + status.count + status.drop == status.total:
            finished = True
        else:
            page += 1
            offset += limit
            requestObj.setOffset(offset)

        # check if cancelled
        proc = open(procFilename)
        cancelled = (proc.read() == 'cancel')
        proc.close()
        if cancelled:
            log("Archiving Cancelled.")

    if not cancelled:
        log("End of dataset. %s out of %s records archived (%s dropped)." % fmt(recordsArchived, status.total, dropped))
        if dropped > 0:
            log("Some records have been dropped due to missing mandatory data items. Please review your source data.", "WARN")
        if dropped < 0:
            log("Some records have been duplicated (see negative number of dropped records); this happens if the source data include\nmultiple values that have been mapped to a non-repeatable element. Please review your source data or modify the mapping.", "WARN")

    # Move zip file to result list
    zipFile.close()
    if not cancelled:
        if recordsArchived == 0:
            # No records found? Remove empty zipfile, add a message to the log, do not return archive filename (because it's empty)
            log('No data to be archived.', 'WARN')
            os.remove(archiveFilenameFull)
            return None, recordsArchived, status.total, dropped
        else:
            # Return archive filename, dataset title and some metadata in case of success
            stat = os.stat(archiveFilenameFull)
            return archiveFilename, recordsArchived, status.total, dropped, stat.st_size, stat.st_mtime

def create_meta(imagefile_exists, identificationfile_exists):
    CORE = '''
    <core encoding="UTF-8" fieldsTerminatedBy="," linesTerminatedBy="\\n" fieldsEnclosedBy=\'"\' ignoreHeaderLines="1" rowType="http://rs.tdwg.org/dwc/terms/Occurrence">
        <files>
            <location>occurrence.txt</location>
        </files>
        <id index="0" />
        <!-- Occurrence fields -->
        <field index="0" term="http://rs.tdwg.org/dwc/terms/catalogNumber"/>
        <field index="1" term="http://rs.tdwg.org/dwc/terms/institutionCode"/>
        <field index="2" term="http://rs.tdwg.org/dwc/terms/collectionCode"/>
        <field index="3" term="http://rs.tdwg.org/dwc/terms/basisOfRecord"/>
        <field index="4" term="http://rs.tdwg.org/dwc/terms/occurrenceID"/>
        <field index="5" term="http://rs.tdwg.org/dwc/terms/fieldNumber"/>
        <field index="6" term="http://purl.org/dc/terms/modified"/>
        <field index="7" term="http://rs.tdwg.org/dwc/terms/eventID"/>
        <field index="8" term="http://rs.tdwg.org/dwc/terms/samplingProtocol"/>
        <field index="9" term="http://rs.tdwg.org/dwc/terms/habitat"/>
        <field index="10" term="http://rs.tdwg.org/dwc/terms/eventRemarks"/>
        <field index="11" term="http://rs.tdwg.org/dwc/terms/minimumElevationInMeters"/>
        <field index="12" term="http://rs.tdwg.org/dwc/terms/maximumElevationInMeters"/>
        <field index="13" term="http://rs.tdwg.org/dwc/terms/minimumDepthInMeters"/>
        <field index="14" term="http://rs.tdwg.org/dwc/terms/maximumDepthInMeters"/>
        <field index="15" term="http://rs.tdwg.org/dwc/terms/country"/>
        <field index="16" term="http://rs.tdwg.org/dwc/terms/countryCode"/>
        <field index="17" term="http://rs.tdwg.org/dwc/terms/locality"/>
        <field index="18" term="http://rs.tdwg.org/dwc/terms/verbatimLocality"/>
        <field index="19" term="http://rs.tdwg.org/dwc/terms/eventDate"/>
        <field index="20" term="http://rs.tdwg.org/dwc/terms/verbatimEventDate"/>
        <field index="21" term="http://rs.tdwg.org/dwc/terms/eventTime"/>
        <field index="22" term="http://rs.tdwg.org/dwc/terms/startDayOfYear"/>
        <field index="23" term="http://rs.tdwg.org/dwc/terms/endDayOfYear"/>
        <field index="24" term="http://rs.tdwg.org/dwc/terms/occurrenceDetails"/>
        <field index="25" term="http://rs.tdwg.org/dwc/terms/occurrenceRemarks"/>
        <field index="26" term="http://rs.tdwg.org/dwc/terms/sex"/>
        <field index="27" term="http://rs.tdwg.org/dwc/terms/decimalLatitude"/>
        <field index="28" term="http://rs.tdwg.org/dwc/terms/decimalLongitude"/>
        <field index="29" term="http://rs.tdwg.org/dwc/terms/coordinateUncertaintyInMeters"/>
        <field index="30" term="http://rs.tdwg.org/dwc/terms/geodeticDatum"/>
        <field index="31" term="http://rs.tdwg.org/dwc/terms/typeStatus"/>
        <field index="32" term="http://rs.tdwg.org/dwc/terms/stateProvince"/>
        <field index="33" term="http://rs.tdwg.org/dwc/terms/county"/>
        <field index="34" term="http://rs.tdwg.org/dwc/terms/municipality"/>
        <field index="35" term="http://rs.tdwg.org/dwc/terms/continent"/>
        <field index="36" term="http://rs.tdwg.org/dwc/terms/waterBody"/>
        <field index="37" term="http://rs.tdwg.org/dwc/terms/islandGroup"/>
        <field index="38" term="http://rs.tdwg.org/dwc/terms/island"/>
        <field index="39" term="http://rs.tdwg.org/dwc/terms/higherGeography"/>
        <!-- Identification fields -->
        <field index="40" term="http://rs.tdwg.org/dwc/terms/dateIdentified"/>
        <field index="41" term="http://rs.tdwg.org/dwc/terms/identifiedBy"/>
        <field index="42" term="http://rs.tdwg.org/dwc/terms/nomenclaturalCode"/>
        <field index="43" term="http://rs.tdwg.org/dwc/terms/taxonRemarks"/>
        <field index="44" term="http://rs.tdwg.org/dwc/terms/identificationQualifier"/>
        <field index="45" term="http://rs.tdwg.org/dwc/terms/identificationRemarks"/>
        <field index="46" term="http://rs.tdwg.org/dwc/terms/identificationReferences"/>
        <field index="47" term="http://rs.tdwg.org/dwc/terms/scientificName"/>
        <field index="48" term="http://rs.tdwg.org/dwc/terms/scientificNameAuthorship"/>
        <field index="49" term="http://rs.tdwg.org/dwc/terms/higherClassification"/>
        <field index="50" term="http://rs.tdwg.org/dwc/terms/kingdom"/>
        <field index="51" term="http://rs.tdwg.org/dwc/terms/phylum"/>
        <field index="52" term="http://rs.tdwg.org/dwc/terms/class"/>
        <field index="53" term="http://rs.tdwg.org/dwc/terms/order"/>
        <field index="54" term="http://rs.tdwg.org/dwc/terms/family"/>
        <field index="55" term="http://rs.tdwg.org/dwc/terms/genus"/>
        <field index="56" term="http://rs.tdwg.org/dwc/terms/subgenus"/>
        <field index="57" term="http://rs.tdwg.org/dwc/terms/specificEpithet"/>
        <field index="58" term="http://rs.tdwg.org/dwc/terms/infraspecificEpithet"/>
        <field index="59" term="http://rs.tdwg.org/dwc/terms/taxonRank"/>
    </core>'''

    IDENTIFICATION = '''
    <extension encoding="UTF-8" fieldsTerminatedBy="," linesTerminatedBy="\\n" fieldsEnclosedBy=\'"\' ignoreHeaderLines="1" rowType="http://rs.tdwg.org/dwc/terms/Identification">
        <files>
            <location>identification.txt</location>
        </files>
        <coreid index="0" />
        <field index="1" term="http://rs.tdwg.org/dwc/terms/dateIdentified"/>
        <field index="2" term="http://rs.tdwg.org/dwc/terms/identifiedBy"/>
        <field index="3" term="http://rs.tdwg.org/dwc/terms/nomenclaturalCode"/>
        <field index="4" term="http://rs.tdwg.org/dwc/terms/taxonRemarks"/>
        <field index="5" term="http://rs.tdwg.org/dwc/terms/identificationQualifier"/>
        <field index="6" term="http://rs.tdwg.org/dwc/terms/identificationRemarks"/>
        <field index="7" term="http://rs.tdwg.org/dwc/terms/identificationReferences"/>
        <field index="8" term="http://rs.tdwg.org/dwc/terms/scientificName"/>
        <field index="9" term="http://rs.tdwg.org/dwc/terms/scientificNameAuthorship"/>
        <field index="10" term="http://rs.tdwg.org/dwc/terms/higherClassification"/>
        <field index="11" term="http://rs.tdwg.org/dwc/terms/kingdom"/>
        <field index="12" term="http://rs.tdwg.org/dwc/terms/phylum"/>
        <field index="13" term="http://rs.tdwg.org/dwc/terms/class"/>
        <field index="14" term="http://rs.tdwg.org/dwc/terms/order"/>
        <field index="15" term="http://rs.tdwg.org/dwc/terms/family"/>
        <field index="16" term="http://rs.tdwg.org/dwc/terms/genus"/>
        <field index="17" term="http://rs.tdwg.org/dwc/terms/subgenus"/>
        <field index="18" term="http://rs.tdwg.org/dwc/terms/specificEpithet"/>
        <field index="19" term="http://rs.tdwg.org/dwc/terms/infraspecificEpithet"/>
        <field index="20" term="http://rs.tdwg.org/dwc/terms/taxonRank"/>
    </extension>'''

    IMAGE = '''
    <extension encoding="UTF-8" fieldsTerminatedBy="," linesTerminatedBy="\\n" fieldsEnclosedBy=\'"\' ignoreHeaderLines="1" rowType="http://rs.gbif.org/terms/1.0/Image">
        <files>
            <location>image.txt</location>
        </files>
        <coreid index="0" />
        <field index="1" term="http://purl.org/dc/terms/identifier"/>
        <field index="2" term="http://purl.org/dc/terms/description"/>
        <field index="3" term="http://purl.org/dc/terms/format"/>
        <field index="4" term="http://purl.org/dc/terms/created"/>
        <field index="5" term="http://purl.org/dc/terms/creator"/>
        <field index="6" term="http://purl.org/dc/terms/license"/>
        <field index="7" term="http://purl.org/dc/terms/rightsHolder"/>
    </extension>'''

    return '<?xml version="1.0" encoding="UTF-8"?>\n<archive xmlns="http://rs.tdwg.org/dwc/text/">%s%s%s\n</archive>' % \
        (CORE, IDENTIFICATION if identificationfile_exists else '', IMAGE if imagefile_exists else '')

def do_convert(archive):
    """Convert the given archive from XML to DwCA"""
    log('Transforming <em>%s</em>' % (archive))

    # Purge/create tmp folder
    try:
        if exists(tmpPath):
            for name in os.listdir(tmpPath):
                os.remove(join(tmpPath, name))
        else:
            os.mkdir(tmpPath)
    except Exception, ex:
        raise Exception("Temporary folder could not be create or purged:\n%s" % (str(ex)))

    # Extract files
    try:
        log("Extracting XML documents...")
        z = zipfile.ZipFile(join (downloadPath, archive), 'r')
        fcount = 0
        for name in z.namelist():
            fcount += 1
            f = open(join(tmpPath, name), 'w')
            f.write(z.read(name))
            f.close()
        z.close()
        # set log level to Minimal for small archives (less than 100 files) and Basic otherwise (more output)
        if fcount > 100:
            level = "Basic"
            log("Archive contains more than 100 documents; showing progress details.")
        else:
            level = "Error"  # other option: "Minimal"
    except Exception, ex:
        raise Exception("XML files could not be extracted from archive: %s." % (str(ex)))

    # Start java process
    try:
        log("Starting transformation (Java heap space: %sM, sort size: %s)..." % (cfg.server.java_mem, cfg.server.sort_size))
        args = ['%s' % (cfg.server.java), '-Djava.ext.dirs=%s' % (cfg.kettleLibLocator),
                '-Djava.io.tmpdir=%s' % (tmpPath), '-DKETTLE_HOME=%s' % (cfg.archiveWorkLocator), '-Xmx%sm' % (cfg.server.java_mem),
                'org.pentaho.di.kitchen.Kitchen', '-level=%s' % (level), '-file=%s' % (join(cfg.archiveLibLocator, 'dwca.kjb')),
                '-param:sort_size=%s' % (cfg.server.sort_size), '-param:base_dir=%s' % (tmpPath), '-param:transformer=%s' % (transformer)]
        l = open(logFile, 'a')
        l.seek(getsize(logFile))
        process = subprocess.Popen(args, shell=False, stdout=l, stderr=l)
        l.close
    except Exception, ex:
        raise Exception('Java process could not be started: %s\nCheck the Test libs page to make sure Java is properly installed and configured.' % (str(ex)))

    # Then, write pid to file
    proc = open(proc_dwca, 'w')
    proc.write(str(process.pid))
    proc.close()

    # Waiiiiiiiiiiiiiiiiit
    ret = process.wait()
    if ret == 0:
        if exists(join(tmpPath, 'occurrence.txt')):
            log('Zipping up files...')
            dwcName = splitext(archive)[0][:-(len(schema))] + DWCA_SCHEMA_NAME + '.zip'
            dwcNameFull = join(archivePath, dwcName)
            f = zipfile.ZipFile(dwcNameFull, 'w', zipfile.ZIP_DEFLATED, True)
            f.writestr('meta.xml', create_meta(exists(join(tmpPath, 'image.txt')), exists(join(tmpPath, 'identification.txt'))))
            for fname in ['occurrence.txt', 'image.txt', 'identification.txt', 'meta.xml', 'eml.xml']:
                fname_full = join(tmpPath, fname)
                if exists(fname_full):
                    f.write(fname_full, fname)
            f.close()

            log("Archive <em>%s</em> successfully created." % (dwcName))
            log(None)
            stat = os.stat(dwcNameFull)
            return (dwcName, stat.st_size, stat.st_mtime)

        else:
            log("No data extracted from XML; DarwinCore Archive could not be created.", "WARN")
            log(None)

    else:
        # Check for the existence of a duplicate.txt
        if exists(dupFile):
            try:
                dup = open(dupFile)
                for line in dup:
                    cat_nr = line.rstrip()
                dup.close()
                log('Duplicate catalog number "%s" found in the dataset. Please eliminate it, recreate the XML archive and retry.' % (cat_nr), 'ERROR')
            except Exception, err:
                log('Duplicate catalog number file could not be opened: ' + str(err), 'WARN')

        # Add some failure message.
        log('Transformation failed with return code %s.\nArchive <em>%s</em> was not converted.' % (ret, archive), "WARN")
        log(None)

        # Transformation cancelled: raise exception
        if ret == 15:
            raise Exception('Conversion cancelled.')



############################################################################################################
### Main archiving functions
############################################################################################################

def create_xml():
    """ Do XML archiving """
    ret_code = -1
    archived, total, dropped = 0, 0, 0
    datasets_empty = 0
    start = time.clock()
    use_filter = True

    try:
        # First, create proc file
        procFilename = join(archivePath, schema + ".proc")
        proc = open(procFilename, "w")
        proc.write("Pid: " + str(os.getpid()))
        proc.write("Process started " + datetime.now().ctime())
        proc.close()

        # Load configuration from file
        configFile = join(archivePath, schema + ".config.xml")
        dom = xml.dom.minidom.parse(configFile)
        schemans = str(dom.getElementsByTagName("schemans")[0].firstChild.data)
        limit = int(dom.getElementsByTagName("limit")[0].firstChild.data)
        log("Archiving schema <i>%s.</i>" % (schemans))
        log("Paging size: %i." % (limit))

        # Remove old archives from working folder
        drop_archives(dsa, False, schema)

        # create operations object
        log("Creating database connection...")
        operationObj = getOperationsObjectForDBMod(dsaObj.getPSFObj())

        # Dataset-aware archiving?
        archives = {}
        if schemans in [NS_ABCD206, NS_HISPID5]:
            # If no dataset list has been passed, get dataset list from inventory
            if datasets:
                log("Using dataset-aware archiving for the requested datasets.")
                xml_datasets = datasets
            else:
                log("Using dataset-aware archiving for this schema. Retrieving dataset list...")
                invObj = InventoryProtocolClass(dsaObj, None)
                invObj.parseRequest(schemans)
                invObj.processRequest()
                xml_datasets = invObj.datasets
                if xml_datasets == []:
                    log("Dataset list could not be retrieved or is empty. \nPlease make sure that the DB connection is OK and that this mapping returns any results!", "ERROR")
                else:
                    msg = "%i dataset%s found." % (len(xml_datasets), '' if len(xml_datasets) == 1 else 's')
                    if len(xml_datasets) == 1:
                        msg += " Using no filter."
                        use_filter = False
                    log(msg)

            #log(str(xml_datasets), "DEBUG")

            # Loop through
            for dataset in xml_datasets:
                # Only use filter if there is more than one dataset (for performance reasons)
                result = do_archive(operationObj, limit, schemans, dataset, use_filter)
                if result is None:
                    # Cancelled: Erase result list, don't copy anything
                    archives = {}
                    break
                else:
                    archived += result[1]
                    total += result[2]
                    dropped += result[3]
                    # Only add archive filename to result list if something has been added
                    if result[0] is None:
                        datasets_empty += 1
                    else:
                        archives[result[0]] = (schema, schemans, dataset, result[1], result[3], result[4], result[5], None)
            log(None)
            if archives != {}:
                log("Done for all datasets. %s out of %s records archived (%s dropped)." % fmt(archived, total, dropped))

        else:
            log("Using dataset-unaware archiving for this schema.")
            result = do_archive(operationObj, limit, schemans)
            if result is not None:
                archived += result[1]
                total += result[2]
                dropped += result[3]
                if result[0] is not None:
                    archives[result[0]] = (schema, schemans, None, result[1], result[3], result[4], result[5], None)
                log(None)

        # If datasets have been skipped, issue a warning
        if datasets_empty != 0:
            log("For some datasets, no archive has been created; please review the log.", "WARN")

        # Otherwise, if no errors occurred: Copy files to download folder
        if archives == {}:
            log("No archives created.", "ERROR")
        else:
            try:
                log("Removing old archives and moving new files to download folder...")
                drop_archives(dsa, True, archives.keys() if datasets else schema)
                for aname in archives.iterkeys():
                    shutil.move(join(archivePath, aname), downloadPath)

                log("Updating inventory...")
                inventory_add(dsa, archives)

                # If XML archiving has been restricted to only certain datasets, only convert theses archives
                if datasets:
                    global filenames
                    filenames = archives.keys()

                # Last, set the status
                if dropped == 0:
                    log('<font color="#009900">SUCCESS: Archiving finished.</font>')
                else:
                    log('<font color="#FF6600">SUCCESS: Archiving finished with warnings.</font>')
                shutil.copy(logFile, downloadPath)
                ret_code = 0

            except Exception, err:
                log("Archive could not be moved to the download folder or error updating the inventory:\n%s" % (str(err)), 'ERROR')


    except Exception, e:
        try:
            log(e.msg if len(e.msg) <= 500 else e.msg[:500] + ' ...', "ERROR")
        except:
            log(str(e), "ERROR")

    finally:
        # If an error occurred, append Fail message
        if ret_code == -1:
            log('Archiving failed!', 'ERROR')

        # If we came here very fast, sleep a bit
        elapsed = time.clock() - start
        if elapsed < 2:
            time.sleep(2 - elapsed)
        if os.path.exists(procFilename):
            os.remove(procFilename)

        # Last, return ret_code
        return ret_code

def create_dwca():
    """Do DwC Archiving"""

    global logFile
    logFile = join(archivePath, DWCA_SCHEMA_NAME + ".log")
    start = time.clock()

    try:
        # First, create proc file
        open(proc_dwca, 'w').close()

        # Remove old archives from working folder
        drop_archives(dsa, False, DWCA_SCHEMA_NAME)

        # Get archive list (or use filename list)
        xml_archives = filenames if filenames else list_archives_for_schema(dsa, schema)
        if xml_archives == []:
            log("No archives found to be converted.", "ERROR")

        # Loop through XML archives
        archives_failed, dwc_archives = 0, {}
        for archive in xml_archives:
            dwc = do_convert(archive)
            if dwc is None:
                archives_failed += 1
            else:
                dwc_archives[dwc[0]] = (DWCA_SCHEMA_NAME, None, None, None, None, dwc[1], dwc[2], archive)

        if dwc_archives == {}:
            log("No archives converted.", "ERROR")
        else:
            # Remove/move DwC Archives
            log("Removing old archives and moving new files to download folder...")
            # If only a single dataset has been converted, remove only this single old archive; otherwise, remove all old archives
            drop_archives(dsa, True, dwc_archives.keys() if filenames else DWCA_SCHEMA_NAME)
            for aname in dwc_archives.iterkeys():
                shutil.move(join(archivePath, aname), downloadPath)

            log("Updating inventory...")
            inventory_add(dsa, dwc_archives)

            # Success message, copy log
            if archives_failed == 0:
                log('<font color="#009900">SUCCESS: Conversion finished.</font>')
            else:
                log('%i archive(s) could not be converted.' % (archives_failed), "WARN")
                log('<font color="#009900">SUCCESS: Conversion finished with warnings.</font>')
            shutil.copy(logFile, downloadPath)


    except Exception, err:
        log(str(err) + "\nDarwinCore Archive generation failed." , "ERROR")

    finally:
        # remove tmp folder
        try:
            if exists(tmpPath):
                shutil.rmtree(tmpPath)
        except Exception, ex:
            log("Temporary folder could not be removed: %s." % (str(ex)), "WARN")

        # If we came here in less than five seconds, we need to wait to make sure that the log is properly displayed in the UI
        elapsed = time.clock() - start
        if elapsed < 5:
            time.sleep(5 - elapsed)

        # remove proc file
        os.remove(proc_dwca)


############################################################################################################
### Main
############################################################################################################

if xmla == 'True':
    ret = create_xml()
    # If both types should be created, only start trnasformation if XML archiving was successful
    if dwca == 'True':
        if ret == 0:
            create_dwca()
else:
    create_dwca()
